#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
SPECEX - SPECtra EXtractor.

This module provides utility to identify sources directly from datacubes.

Copyright (C) 2022-2023  Maurizio D'Addona <mauritiusdadd@gmail.com>
"""
from __future__ import absolute_import, division, print_function

import os
import sys
import argparse
import numpy as np
import multiprocessing as mp

import matplotlib.pyplot as plt

from astropy import wcs
from astropy.io import fits
from astropy.table import Table
from astropy.convolution import Gaussian2DKernel, convolve

from scipy.ndimage.morphology import grey_dilation  # , grey_erosion

from specex.utils import get_pbar
from specex.utils import get_spectrum_snr, get_spectrum_snr_emission
from specex.cube import get_hdu


def __argshandler(options=None):
    """
    Parse the arguments given by the user.

    Returns
    -------
    args: Namespace
        A Namespace containing the parsed arguments. For more information see
        the python documentation of the argparse module.
    """
    parser = argparse.ArgumentParser(
        description='Identify sources directly from the datacube.'
    )
    parser.add_argument(
        'input_cube', metavar='SPEC_CUBE', type=str, nargs=1,
        help='The spectral cube in fits format from which spectra will be '
        'extracted.'
    )

    parser.add_argument(
        '--out-regionfile', metavar='REGFILE', type=str, default=None,
        help='The path of a region file where to save identified sources'
    )

    parser.add_argument(
        '--sn-threshold', metavar='SN_THRESH', type=float, default=1.5,
        help='Set the signal to Noise ratio threshold: objects with their '
        'spectrum having a SN ratio lower than threshold will be ignored. '
        'If SN_THRESH is set to a value equal or lower than 0, SN filtering '
        'is disabled'
    )

    parser.add_argument(
        '--info-hdu', metavar='INFO_HDU', type=int, default=0,
        help='The HDU containing cube metadata. If this argument '
        'Set this to -1 to automatically detect the HDU containing the info. '
        'NOTE that this value is zero indexed (i.e. firts HDU has index 0).'
    )

    parser.add_argument(
        '--spec-hdu', metavar='SPEC_HDU', type=int, default=-1,
        help='The HDU containing the spectral data to use. If this argument '
        'Set this to -1 to automatically detect the HDU containing spectra. '
        'NOTE that this value is zero indexed (i.e. second HDU has index 1).'
    )

    parser.add_argument(
        '--var-hdu', metavar='VAR_HDU', type=int, default=-1,
        help='The HDU containing the variance of the spectral data. '
        'Set this to -1 if no variance data is present in the cube. '
        'The default value is %(metavar)s=%(default)s.'
        'NOTE that this value is zero indexed (i.e. third HDU has index 2).'
    )

    parser.add_argument(
        '--mask-hdu', metavar='MASK_HDU', type=int, default=-1,
        help='The HDU containing the valid pixel mask of the spectral data. '
        'Set this to -1 if no mask is present in the cube. '
        'The default value is %(metavar)s=%(default)s.'
        'NOTE that this value is zero indexed (i.e. fourth HDU has index 3).'
    )

    parser.add_argument(
        '--invert-mask', action='store_true', default=False,
        help='Set whether to use the inverse of the cube mask.'
    )

    parser.add_argument(
        '--outdir', metavar='DIR', type=str, default=None,
        help='Set the directory where extracted spectra will be outputed. '
        'If this parameter is not specified, then a new directory will be '
        'created based on the name of the input cube.'
    )

    parser.add_argument(
        '--check-images', action='store_true', default=False,
        help='Whether or not to generate check images.'
    )

    parser.add_argument(
        "--checkimg-outdir", type=str, default=None, required=False,
        help='Set the directory where check images are saved (when they are '
        'enabled thorugh the appropriate parameter).'
    )

    args = None
    if options is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(options)

    return args


def rebin_slice_2d(data, bin_size_x, bin_size_y=None, rebin_function=np.mean):
    """
    Rebin a 2D image.

    Parameters
    ----------
    data : 2D np.ndarray
        The input array.
    bin_size_x : int
        Number of pixels to combine along x axis.
    bin_size_y : int, optional
        Number of pixels to combine along x axis. If none, use the same value
        as bin_size_x The default is None.
    rebin_function : function, optional
        The function used to combine pixels. The default is np.mean.

    Returns
    -------
    rebinned : 2D np.ndarray
        The rebinned array.

    """
    bin_size_x = int(bin_size_x)

    if bin_size_y is None:
        bin_size_y = bin_size_x
    else:
        bin_size_y = int(bin_size_y)

    if bin_size_x <= 0:
        raise ValueError("bin_size_x should be greater than zero!")

    if bin_size_y <= 0:
        raise ValueError("bin_size_y should be greater than zero!")

    if bin_size_x == 1 and bin_size_y == 1:
        raise ValueError(
            "At least one between bin_size_x and bin_size_y should be greater"
            " than one!"
        )

    rebinned_h = data.shape[0] // bin_size_y
    rebinned_w = data.shape[1] // bin_size_x

    # Resize data to optimal height and width for rebinning.
    if data.shape[0] % bin_size_y:
        rebinned_h += 1
        optimal_data_height = rebinned_h * bin_size_y
    else:
        optimal_data_height = data.shape[0]

    if data.shape[1] % bin_size_x:
        rebinned_w += 1
        optimal_data_width = rebinned_w * bin_size_x
    else:
        optimal_data_width = data.shape[1]

    # Padding with original array in order to reshape it correctly
    zeros = np.zeros((optimal_data_height, optimal_data_width))
    masked_data = np.ma.array(zeros, mask=zeros == 0)
    masked_data.mask[:data.shape[0], :data.shape[1]] = False
    masked_data[:data.shape[0], :data.shape[1]] = data

    # Reshape the array
    reshaped = masked_data.reshape(
        rebinned_h, bin_size_y,
        rebinned_w, bin_size_x
    ).transpose(0, 2, 1, 3)

    reshaped = reshaped.reshape(rebinned_h, rebinned_w, bin_size_y*bin_size_x)

    # Do actual rebinning
    rebinned = rebin_function(reshaped, axis=-1)

    return rebinned


def get_sn_single_slice(cube_slice, cube_slice_var=None, rebin_size=2,
                        kernel_width=1):
    """
    Get Signal to Noise ratio of an image.

    Identify spotential ources from a single 2D array (image) considering the
    signal to noise ratio of each pixel. If a pixel has a SN greather than the
    given threshold then it is considered to belong to a source.

    NOTE
    ----
    This function is *very* memory intensive and will probably never be
    actually used since I wrote more memory-wise functions. I leave it
    here just as an example.

    Parameters
    ----------
    cube_slice : 2D np.ndarray
        A single slice of the datacube.
    cube_slice_var : 2D np.ndarray, optional
        The variance map for the input slice. If this is not provided,
        noise will be approximately estimated directly from the input slice.
    noise_rebin_size : int, optional
        Both imput slice and  cube_slice_var (if provided) are rebinned
        combining NxN pixel into one pixel, where N = 2*rebin_size + 1.
    kernel_width : float:
        The std. dev. to be passed to the smoothing gaussian kernel.
        (For more info see astropy.convolution.Gaussian2DKernel)

    Returns
    -------
    None.

    """
    # Compute the SN ratio for each pixel in the slice.
    # First of all, smooth the image using a 2D gaussian kernel.
    kernel = Gaussian2DKernel(x_stddev=kernel_width)  # x_stddev = y_stddev
    smoothed_slice = convolve(
        cube_slice,
        kernel,
        nan_treatment='fill',
        fill_value='-1',
        preserve_nan=True
    )

    bin_size = 2 * rebin_size + 1

    if cube_slice_var is None:
        # No variance information, try to perform a crude estimation of the
        # noise by subtracting the smoothed image from the original image
        noise_image = cube_slice - smoothed_slice

        # Rebinning to get variance
        rebinned_var = rebin_slice_2d(
            noise_image, bin_size, rebin_function=np.var
        )

        # Rescale back to original size
        cube_slice_var = np.kron(rebinned_var, np.ones((bin_size, bin_size)))
        cube_slice_var = cube_slice_var[
            :cube_slice.shape[0],
            :cube_slice.shape[1]
        ]

    sn_map = smoothed_slice / np.sqrt(cube_slice_var)

    return sn_map


def get_snr_spaxels(data_hdu, var_hdu=None, mask_hdu=None, inverse_mask=False,
                    snr_function=get_spectrum_snr):
    """
    Get the SNR map of a spectral datacube.

    Parameters
    ----------
    data_hdu : astropy.fits.ImageHDU
        The HDU containing spectral data.
    mask_hdu : astropy.fits.ImageHDU, optional
        The HDU containing data qaulity mask. The default is None.
    inverse_mask : bool, optional
        Wether. The default is False.

    Returns
    -------
    snr_map : TYPE
        DESCRIPTION.

    """
    cube_mask = np.isnan(data_hdu.data)
    if mask_hdu is not None:
        cube_mask &= mask_hdu.data != 0 if inverse_mask else mask_hdu.data == 0

    if var_hdu is not None:
        cube_var = np.ma.array(
            var_hdu.data, mask=cube_mask & np.isnan(var_hdu.data)
        )
    else:
        cube_var = None

    cube_flux = np.ma.array(data_hdu.data, mask=cube_mask)

    snr_map = np.zeros((cube_flux.shape[1], cube_flux.shape[2]))

    i = 0
    footprint_size = cube_flux.shape[1] * cube_flux.shape[2]
    results = {}
    # Generating jobs...
    with mp.Pool() as my_pool:
        for h in range(cube_flux.shape[1]):
            for k in range(cube_flux.shape[2]):
                i += 1
                if (i % 100) == 0:
                    progress = 0.5 * i / footprint_size
                    sys.stderr.write(
                        f"\r{get_pbar(progress)} {progress:.2%}\r"
                    )
                    sys.stderr.flush()
                # The snr of a single spaxel
                # snr_map[h, k] = get_spectrum_snr(cube_flux[:, h, k])

                if cube_flux[:, h, k].mask.all():
                    continue

                results[h, k] = my_pool.apply_async(
                    snr_function,
                    (
                        cube_flux[:, h, k],
                        cube_var[:, h, k] if cube_var is not None else None
                    ),
                )

        for i, (pos, val) in enumerate(results.items()):
            if (i % 100) == 0:
                progress = 0.5 + 0.5 * i / footprint_size
                sys.stderr.write(
                    f"\r{get_pbar(progress)} {progress:.2%}\r"
                )
                sys.stderr.flush()
            val.wait()
            snr_map[pos] = val.get()
    print("")
    return snr_map


def get_local_maxima(image, threshold=0, smoothing_factor=1):
    """
    Find local peaks in an image comparing pixels with their neighbours.

    Parameters
    ----------
    image : 2D numpy.ndarray
        The image data.
    threshold : float, optional
        Peaks with value lower or equal than this value will be discarded.
    smoothing_factor : float, optional
        Std. Dev. of the gaussian kernel used to smooth the input image.

    Returns
    -------
    peaks : list
        list of 2-tuples containing the positions of the local maxima.
    """
    hh, ww = image.shape
    p_map = np.zeros_like(image)

    smoothing_kernel = Gaussian2DKernel(x_stddev=1)
    im_smooth = convolve(
        image,
        smoothing_kernel,
        preserve_nan=True
    )

    for h in range(1, hh-1):
        for k in range(1, ww-1):
            if (
                im_smooth[h, k] > im_smooth[h-1, k] and
                im_smooth[h, k] > im_smooth[h+1, k] and
                im_smooth[h, k] > im_smooth[h, k-1] and
                im_smooth[h, k] > im_smooth[h, k+1]
            ):
                p_map[h, k] += 1

    peaks = np.argwhere(p_map > 0)

    valid_peaks = np.array(
        [
            (x[0], x[1], im_smooth[x[0], x[1]]) for x in peaks
            if not np.isnan(im_smooth[x[0], x[1]]) and
            im_smooth[x[0], x[1]] > threshold
        ]
    )

    tbl = Table(data=valid_peaks, names=['Y_IMAGE', 'X_IMAGE', 'SNR'])

    return tbl


def get_dilation_maxima(image, box_sizes=[3, 5], threshold=0,
                        smoothing_factor=1):
    """
    Find local peaks in an image using a dilation morphological filter.

    Parameters
    ----------
    image : 2D numpy.ndarray
        The image data.
    box_size : int, optional
        The size of the dilation box. The default is 3.
        See https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.maximum_filter.html
    threshold : float, optional
        Peaks with value lower or equal than this value will be discarded.
    smoothing_factor : float, optional
        Std. Dev. of the gaussian kernel used to smooth the input image.

    Returns
    -------
    peaks : list
        list of 2-tuples containing the positions of the local maxima.
    """
    if not box_sizes:
        box_sizes = [3, 5]

    smoothing_kernel = Gaussian2DKernel(x_stddev=1)
    im_smooth = convolve(
        image,
        smoothing_kernel,
        preserve_nan=True
    )

    dilated = [
        grey_dilation(im_smooth, size=x) for x in box_sizes
    ]

    mask = np.ones_like(im_smooth, dtype=bool)

    for d_img in dilated:
        mask &= im_smooth == d_img

    # TODO: What happens if we have a region of pixels of euqal values?
    #       In this case we should find the center of mass of the region?
    peaks = np.argwhere(mask)

    valid_peaks = np.array(
        [
            (x[0], x[1], im_smooth[x[0], x[1]]) for x in peaks
            if not np.isnan(im_smooth[x[0], x[1]]) and
            im_smooth[x[0], x[1]] > threshold
        ]
    )

    tbl = Table(data=valid_peaks, names=['Y_IMAGE', 'X_IMAGE', 'SNR'])

    return tbl


def get_spectrum(flux_spaxels, var_spaxels=None, weights=None, average=False):
    """
    Get spectrum from a list of spaxels.

    Parameters
    ----------
    flux_spaxels : 2D numpy.ndarray
        The set of spectralspaxels used to compute the spectrum, in the form of
        a 2D numpy array where the first axis represents the wavelength and the
        second axis moves trough the different spaxels.
    var_spaxels : 2D numpy.ndarray, optional
        Same format of flux_spaxels but for variace data. The default is None.
        If none, the variance is computed using the spectra lspaxels.
    weights : 1D numpy.ndarray or list of float, optional
        Optional weights for each spaxel. The default is None.
    average : TYPE, optional
        Wether to average the spaxels or to sum them. The default is False.

    Returns
    -------
    spectrum : 1D numpy.ndarray
        The spectrum.
    spectrum_var :  1D numpy.ndarray
        The variance of the spectrum.
    """
    if var_spaxels is not None:
        if not isinstance(var_spaxels, np.ma.MaskedArray):
            var_spaxels = np.ma.array(
                var_spaxels, mask=~np.isfinite(var_spaxels)
            )

        if var_spaxels.shape != flux_spaxels.shape:
            raise ValueError("Flux and variance have different shape!")

    if not isinstance(flux_spaxels, np.ma.MaskedArray):
        flux_spaxels = np.ma.array(
            flux_spaxels, mask=~np.isfinite(flux_spaxels)
        )

    if weights is None:
        spatial_w = 1
    else:
        if not isinstance(var_spaxels, np.ma.MaskedArray):
            spatial_w = np.ma.array(weights, mask=~np.isfinite(weights))
        else:
            spatial_w = weights
        spatial_w /= np.ma.max(spatial_w)

    # Summing spaxels weighted by spatial_w on the spatial axis

    w_sum = np.ma.sum(spatial_w)
    spectrum = np.ma.sum(flux_spaxels * spatial_w, axis=-1)
    spectrum /= w_sum

    if var_spaxels is None:
        spectrum_var = np.ma.var(flux_spaxels * spatial_w, axis=-1)
    else:
        spectrum_var = np.ma.sum(var_spaxels * spatial_w, axis=-1)
    spectrum_var /= w_sum

    return spectrum, spectrum_var


def source_clustering(sources_table, data_hdu, var_hdu=None, mask_hdu=None,
                      snr_function=get_spectrum_snr, max_distance=3,
                      nmae_threshold=1, inverse_mask=False):

    def has_a_neighbour(arr, y, x):
        val = (arr[y+1, x] != -1) or (arr[y-1, x] != -1)
        val = val or (arr[y, x-1] != -1) or (arr[y, x-1] != -1)
        val = val or (arr[y-1, x-1] != -1) or (arr[y+1, x+1] != -1)
        val = val or (arr[y-1, x+1] != -1) or (arr[y+1, x-1] != -1)
        return val

    if mask_hdu is None:
        cube_mask = np.isnan(data_hdu.data)
    else:
        cube_mask = mask_hdu.data != 0 if inverse_mask else mask_hdu.data == 0

    if var_hdu is not None:
        cube_var = np.ma.array(
            var_hdu.data, mask=cube_mask & np.isnan(var_hdu.data)
        )
    else:
        cube_var = None

    cube_flux = np.ma.array(data_hdu.data, mask=cube_mask)

    source_map = -1 * np.ones((cube_flux.shape[1], cube_flux.shape[2]))

    # source_prob = np.ones((cube_flux.shape[1], cube_flux.shape[2]))

    for sid, source in enumerate(sources_table):
        progress = sid / len(sources_table)
        sys.stderr.write(
            f"\r{get_pbar(progress)} {progress:.2%}\r"
        )
        sys.stderr.flush()

        source_x = int(source['X_IMAGE'])
        source_y = int(source['Y_IMAGE'])

        if source_map[source_y, source_x] >= 0:
            sid = source_map[source_y, source_x]
        else:
            source_map[source_y, source_x] = sid

        base_spectrum = cube_flux[..., source_y, source_x]
        if cube_var is not None:
            base_var = cube_var[..., source_y, source_x]
        else:
            base_var = None

        base_snr = snr_function(base_spectrum, base_var)

        pixels = []
        for h in range(-max_distance, max_distance+1):
            for k in range(-max_distance, max_distance+1):
                if h == 0 and k == 0:
                    # this is base_spectrum itself
                    continue
                elif (h**2 + k**2) > max_distance**2:
                    continue
                new_sp_y = int(source['Y_IMAGE'] + h)
                new_sp_x = int(source['X_IMAGE'] + k)
                pixels.append((new_sp_y, new_sp_x, h**2 + k**2))

        pixels.sort(key=lambda tup: tup[2])

        for j, (new_sp_y, new_sp_x, d) in enumerate(pixels):
            progress = (sid + j/len(pixels)) / len(sources_table)
            if j % 10 == 0:
                sys.stderr.write(
                    f"\r{get_pbar(progress)} {progress:.2%}\r"
                )
                sys.stderr.flush()
            # Add a new pixel to source_map

            try:
                if not has_a_neighbour(source_map, new_sp_y, new_sp_x):
                    continue
            except IndexError:
                continue

            _old_map_value = source_map[new_sp_y, new_sp_x]
            source_map[new_sp_y, new_sp_x] = sid

            # Check wheter we improved the SNR of the total spectrum
            flux_spaxels = cube_flux[:, source_map == sid]

            if cube_var is not None:
                var_spaxels = cube_var[:, source_map == sid]
            else:
                var_spaxels = None

            new_spectrum, new_var = get_spectrum(
                flux_spaxels,
                var_spaxels,
                average=False
            )

            new_snr = snr_function(new_spectrum, new_var)

            # if SNR did not improve then remove this pixel from the map
            if new_snr > base_snr:
                base_snr = new_snr
                source_map[new_sp_y, new_sp_x] = sid
                if _old_map_value != -1:
                    # We want to add this spaxel to the current source but
                    # it already belongs to another source... merging
                    source_map[source_map == _old_map_value] = sid
                    pass
            else:
                source_map[new_sp_y, new_sp_x] = _old_map_value

    return source_map


def detect_from_cube(options=None):
    """
    Run the main program of this module.

    Parameters
    ----------
    options : list, optional
        List of arguments to pass to argparse. The default is None.

    Returns
    -------
    None.

    """
    args = __argshandler(options)

    basename_with_ext = os.path.basename(args.input_cube[0])
    basename = os.path.splitext(basename_with_ext)[0]

    if args.outdir is None:
        outdir = f'{basename}_spectra'
    else:
        outdir = args.outdir

    if not os.path.isdir(outdir):
        os.mkdir(outdir)

    if args.checkimg_outdir is not None:
        check_images_outdir = args.checkimg_outdir
    else:
        check_images_outdir = os.path.join(outdir, 'checkimages')

    if args.check_images and not os.path.isdir(check_images_outdir):
        os.mkdir(check_images_outdir)

    hdl = fits.open(args.input_cube[0])

    spec_hdu = get_hdu(
        hdl,
        hdu_index=args.spec_hdu,
        valid_names=['data', 'spec', 'spectrum', 'spectra'],
        msg_err_notfound="ERROR: Cannot determine which HDU contains spectral "
                         "data, try to specify it manually!",
        msg_index_error="ERROR: Cannot open HDU {} to read specra!"
    )

    var_hdu = get_hdu(
        hdl,
        hdu_index=args.var_hdu,
        valid_names=['stat', 'var', 'variance', 'noise'],
        msg_err_notfound="WARNING: Cannot determine which HDU contains the "
                         "variance data, try to specify it manually!",
        msg_index_error="WARNING: Cannot open HDU {} to read the "
                        "variance!",
        exit_on_errors=False
    )

    mask_hdu = get_hdu(
        hdl,
        hdu_index=args.mask_hdu,
        valid_names=['mask', 'platemask', 'footprint', 'dq'],
        msg_err_notfound="WARNING: Cannot determine which HDU contains the "
                         "mask data, try to specify it manually!",
        msg_index_error="WARNING: Cannot open HDU {} to read the mask!",
        exit_on_errors=False
    )

    cube_wcs = wcs.WCS(spec_hdu.header)
    celestial_wcs = cube_wcs.celestial

    """
    snr_map = get_snr_map(spec_hdu, var_hdu, mask_hdu)
    snr_map = snr_map.filled(fill_value=np.nan)
    hdu = fits.PrimaryHDU(data=snr_map, header=celestial_wcs.to_header())
    hdu.writeto(f"{basename}_snr_map.fits", overwrite=True)
    """

    print("\nComputing SNR of emission fearures...", file=sys.stderr)
    snr_map_em = get_snr_spaxels(
        spec_hdu, var_hdu, mask_hdu,
        snr_function=get_spectrum_snr_emission
    )
    hdu = fits.PrimaryHDU(data=snr_map_em, header=celestial_wcs.to_header())
    hdu.writeto(
        os.path.join(outdir, f"{basename}_snr_map_em.fits"),
        overwrite=True
    )

    if args.check_images:
        fig, ax = plt.subplots(1, 1, figsize=(10, 10))
        ax.imsho()
        fig.savefig(
            os.path.join(check_images_outdir, f"{basename}_snr_map_em.fits")
        )
        plt.tight_layout()

    print("\nComputing SNR of the continuum...", file=sys.stderr)
    snr_map_ct = get_snr_spaxels(
        spec_hdu, var_hdu, mask_hdu,
        snr_function=get_spectrum_snr
    )
    hdu = fits.PrimaryHDU(data=snr_map_ct, header=celestial_wcs.to_header())
    hdu.writeto(
        os.path.join(outdir, f"{basename}_snr_map_cont.fits"),
        overwrite=True
    )

    smoothing_kernel = Gaussian2DKernel(x_stddev=1.5)

    snr_map_em_smooth = convolve(
        snr_map_em,
        smoothing_kernel,
        preserve_nan=True
    )
    snr_map_ct_smooth = convolve(
        snr_map_ct,
        smoothing_kernel,
        preserve_nan=True
    )

    snr_map_em_smooth[snr_map_em_smooth < 0] = 0
    snr_map_ct_smooth[snr_map_ct_smooth < 0] = 0

    snr_map = snr_map_em_smooth + snr_map_ct_smooth
    hdu = fits.PrimaryHDU(data=snr_map, header=celestial_wcs.to_header())
    hdu.writeto(f"{basename}_snr_map_total.fits", overwrite=True)

    print("\nDetecting local maxima...", file=sys.stderr)
    peaks = get_local_maxima(snr_map_ct, threshold=0.1)

    print("\nImage segmentation...", file=sys.stderr)
    smap = source_clustering(peaks, spec_hdu, var_hdu, max_distance=3)
    hdu = fits.PrimaryHDU(data=smap, header=celestial_wcs.to_header())
    hdu.writeto(f"{basename}_segmentation.fits", overwrite=True)


if __name__ == '__main__':
    detect_from_cube()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
SPECEX - SPECtra EXtractor.

This module provides utility functions and executable to plot spectra.

Copyright (C) 2022-2023  Maurizio D'Addona <mauritiusdadd@gmail.com>
"""
import os
import sys
import argparse
import json
import tempfile
import subprocess

from glob import glob

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec

from scipy.ndimage import gaussian_filter

from astropy import units as apu
from astropy.io import fits
from astropy import wcs
from astropy.table import Table, MaskedColumn
from astropy.coordinates import SkyCoord
from astropy.nddata import Cutout2D
from astropy.visualization import quantity_support

from specex.utils import plot_spectrum, get_pbar, load_rgb_fits, find_prog
from specex.cube import get_hdu, get_rgb_cutout, get_gray_cutout

from specex.cubex import load_specex_file

try:
    import imageio
except ModuleNotFoundError:
    HAS_IMAGEIO = False
else:
    HAS_IMAGEIO = True

FFMPEG_EXC = find_prog('ffmpeg')
if FFMPEG_EXC is None:
    FFMPEG_EXC = find_prog('ffmpeg.exe')


def __plot_spectra_argshandler(options=None):
    """
    Parse the arguments given by the user.

    Returns
    -------
    args: Namespace
        A Namespace containing the parsed arguments. For more information see
        the python documentation of the argparse module.
    """
    parser = argparse.ArgumentParser(
        description='Plot spectra extracted with specex.'
    )

    parser.add_argument(
        '--outdir', metavar='DIR', type=str, default=None,
        help='Set the directory where extracted spectra will be outputed. '
        'If this parameter is not specified, then plots will be saved in the '
        'same directory of the corresponding input spectrum file.'
    )

    parser.add_argument(
        '--zcat', metavar='Z_CAT_FILE', type=str, default=None,
        help='If specified then the catalog %(metavar)s is used to read the '
        'redshift of the spectra. The catalog must contain at least two '
        'columns: one for the id of the objects and one for the redshifts. '
        'The name of the columns can be set using the parameters --key-id and '
        '--key-z. When this option is used, spectra having no matching ID in '
        'the catalog are skipped.'
    )

    parser.add_argument(
        '--key-id', metavar='KEY_ID', type=str, default='SPECID',
        help='Set the name of the column in the zcat that contains the IDs of '
        'the spectra. See --zcat for more info. If this option is not '
        'specified, then he default value %(metavar)s = %(default)s is used.'
    )

    parser.add_argument(
        '--key-z', metavar='KEY_Z', type=str, default='Z',
        help='Set the name of the column in the zcat that contains the '
        'redshift the spectra. See --zcat for more info. If this option is not'
        ' specified, then he default value %(metavar)s = %(default)s is used.'
    )

    parser.add_argument(
        '--key-wrange', metavar='WAVE_RANGE', type=str, default=None,
        help='Set the name of the column in the zcat that contains the range'
        'of wavelength to plot. If not specified then the whole spectrum is '
        'plotted. If specified and the range value is empty no plot is '
        'produced for the object.'
    )

    parser.add_argument(
        '--restframe', default=False, action='store_true',
        help='If this option is specified, spectra will be plotted as if they '
        'were in the observer restframe (ie. they are de-redshifted). '
        'In order to use this option, a zcat must be specified.'
    )

    parser.add_argument(
        '--smoothing', metavar='WINDOW_SIZE', type=int,  default=3,
        help='If %(metavar)s >= 0, then plot a smoothed version of the '
        'spectrum alongside the original one.  %(metavar)s = 0 means that '
        'only the original spectrum is plotted. If not specified, the default '
        '%(metavar)s = %(default)s is used.'
    )

    parser.add_argument(
        '--cutout', metavar='SOURCE_IMAGE', type=str, default=None,
        help='path of a FITS image (RGB or grayscale) used to make cutouts '
        'that will be included in the plots. The size of the cutout can be '
        'set using the --cutout-size option.'
    )

    parser.add_argument(
        '--cutout-size', metavar='SIZE', type=str, default='2arcsec',
        help='Set the size of the cutout to %(metavar)s. This option '
        'supports units compatible with astropy (for example "1deg", '
        '"2arcmin", "5arcsec", etc.). If no unit is specified the size is '
        'assumed to be in arcseconds. The default cutout size is %(default)s.'
    )

    parser.add_argument(
        'spectra', metavar='SPECTRUM', type=str, nargs='+',
        help='Input spectra extracted with specex.'
    )

    args = None
    if options is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(options)

    return args


def plot_spectra(options=None):
    """
    Run the plot program.

    Returns
    -------
    None.

    """
    quantity_support()
    args = __plot_spectra_argshandler(options)

    if args.zcat is not None:
        zcat = Table.read(args.zcat)
        if args.key_id not in zcat.colnames:
            print(
                f"ERROR: z catalog does not have id column '{args.key_id}'",
                file=sys.stderr
            )
            sys.exit(1)
        elif args.key_z not in zcat.colnames:
            print(
                f"ERROR: z catalog does not have z column '{args.key_z}'",
                file=sys.stderr
            )
            sys.exit(1)

        # Remove objects with masked or undefined IDs
        if isinstance(zcat[args.key_id], MaskedColumn):
            zcat = zcat[~zcat[args.key_id].mask]

        zcat.add_index(args.key_id)
    else:
        zcat = None

    if args.cutout is not None:
        try:
            big_image = load_rgb_fits(args.cutout)
        except FileNotFoundError:
            print(f"ERROR: file not found '{args.cutout}'")
            sys.exit(1)

        if big_image is None:
            big_image = {
                'data': fits.getdata(args.cutout),
                'wcs': wcs.WCS(fits.getheader(args.cutout)),
                'type': 'gray'
            }

        cutout_size = apu.Quantity(args.cutout_size)
    else:
        big_image = None

    # Windows prompt does not expand globs, so let's do it
    spectra_list = []
    for globbed_fname in args.spectra:
        for fname in glob(globbed_fname):
            spectra_list.append(fname)

    for j, spectrum_fits_file in enumerate(spectra_list):
        progress = j / len(spectra_list)
        sys.stdout.write(f"\r{get_pbar(progress)} {progress:.2%}\r")
        sys.stdout.flush()

        sp_dict = load_specex_file(spectrum_fits_file)

        if any(
                x is None for x in [
                    sp_dict['main_header'],
                    sp_dict['flux'],
                    sp_dict['variance']
                ]
        ):
            print(f"Skipping file '{spectrum_fits_file}'\n")
            continue

        try:
            object_ra = sp_dict['main_header']['RA']
            object_dec = sp_dict['main_header']['DEC']
            object_id = sp_dict['main_header']['ID']
            extraction_mode = sp_dict['main_header']['EXT_MODE']
            specex_apertures = [
                apu.Quantity(x)
                for x in json.loads(sp_dict['main_header']['EXT_APER'])
            ]
        except KeyError:
            print(
                f"Skipping file with invalid header: {spectrum_fits_file}"
            )
            continue
        else:

            try:
                object_coord_frame = sp_dict['main_header']['FRAME']
            except KeyError:
                object_coord_frame = 'icrs'

            obj_center = SkyCoord(
                object_ra, object_dec,
                unit='deg',
                frame=object_coord_frame
            )

        info_dict = {
            'ID': f"{object_id}",
            'RA': obj_center.ra.to_string(precision=2),
            'DEC': obj_center.dec.to_string(precision=2),
            'FRAME': str(object_coord_frame).upper()
        }
        for key in ['Z', 'SN', 'SN_EMISS']:
            try:
                info_dict[key] = f"{sp_dict['main_header'][key]:.4f}"
            except KeyError:
                continue

        wave_range = None
        if zcat is not None:
            try:
                object_z = zcat.loc[object_id][args.key_z]
            except KeyError:
                print(
                    f"WARNING: '{object_id}' not in zcat, skipping...",
                    file=sys.stderr
                )
                continue
            else:
                try:
                    # In case of repeated objects
                    object_z = object_z[0]
                except IndexError:
                    # Otherwise just go ahead
                    pass

                if args.key_wrange is not None:
                    str_wrange = zcat.loc[object_id][args.key_wrange]
                    try:
                        wave_range = [
                            float(x) for x in str_wrange.split('-')
                        ]
                    except Exception:
                        continue

            restframe = args.restframe
            info_dict['Z'] = object_z
        else:
            # If no zcat is provided, check if redshift information is
            # stored in the spectrum itself
            if 'Z' in info_dict:
                object_z = float(info_dict['Z'])
                restframe = args.restframe
            else:
                object_z = None
                restframe = False

        if big_image is not None:
            if big_image['type'] == 'rgb':
                cutout_dict = get_rgb_cutout(
                    big_image['data'],
                    center=obj_center,
                    size=cutout_size,
                    data_wcs=big_image['wcs']
                )
                cutout = np.asarray(cutout_dict['data']).transpose(1, 2, 0)
                cutout_wcs = cutout_dict['wcs'][0]
            else:
                cutout_dict = get_gray_cutout(
                    big_image['data'],
                    center=obj_center,
                    size=cutout_size,
                    data_wcs=big_image['wcs']
                )
                cutout = np.array(cutout_dict['data'])
                cutout_wcs = cutout_dict['wcs']
            cutout_vmin = np.nanmin(big_image['data'])
            cutout_vmax = np.nanmax(big_image['data'])
        else:
            cutout = None
            cutout_wcs = None
            cutout_vmin = None
            cutout_vmax = None

        fig, axs = plot_spectrum(
            sp_dict['wavelength'],
            sp_dict['flux'],
            sp_dict['variance'],
            nan_mask=sp_dict['nan_mask'],
            redshift=object_z,
            restframe=restframe,
            cutout=cutout,
            cutout_wcs=cutout_wcs,
            cutout_vmin=cutout_vmin,
            cutout_vmax=cutout_vmax,
            flux_units=sp_dict['flux_units'],
            wavelengt_units=sp_dict['wavelength_units'],
            smoothing=args.smoothing,
            extra_info=info_dict,
            extraction_info={
                'mode': extraction_mode,
                'apertures': specex_apertures,
                'aperture_ra': object_ra,
                'aperture_dec': object_dec,
                'frame': object_coord_frame
            },
            wave_range=wave_range
        )

        if args.outdir is None:
            outdir = os.path.dirname(spectrum_fits_file)
        else:
            outdir = args.outdir
            if not os.path.isdir(outdir):
                os.makedirs(outdir)

        fig_out_name = os.path.basename(spectrum_fits_file)
        fig_out_name = os.path.splitext(fig_out_name)[0]
        fig_out_path = os.path.join(outdir, f"{fig_out_name}.png")
        fig.savefig(fig_out_path, dpi=150)
        plt.close(fig)

    print(f"\r{get_pbar(1)} 100%")


def __plot_slice_argshandler(options=None):
    """
    Parse the arguments given by the user.

    Returns
    -------
    args: Namespace
        A Namespace containing the parsed arguments. For more information see
        the python documentation of the argparse module.
    """
    global FFMPEG_EXC

    parser = argparse.ArgumentParser(
        description='Plot spectra extracted with specex.'
    )

    parser.add_argument(
        '--spec-smoothing', metavar='WINDOW_SIZE', type=int,  default=0,
        help='If %(metavar)s >= 0, then plot a smoothed version of the '
        'spectrum alongside the original one.  %(metavar)s = 0 means that '
        'only the original spectrum is plotted. If not specified, the default '
        '%(metavar)s = %(default)s is used.'
    )

    parser.add_argument(
        '--cube-smoothing', metavar='WINDOW_SIZE', type=float,  default=0,
        help='If %(metavar)s >= 0, then plot a smoothed version of the '
        'cube cutout.  %(metavar)s = 0 means that no smoothing is applied.'
        'The default value is %(metavar)s=%(default)s.'
    )

    parser.add_argument(
        '--cube-vlim', metavar='VMIN,VMAX', type=str, default=None,
        help='Set the minimum and maximum values for the colormap of the '
        'cube animated cutout. If not specified, then vmin and vmax are '
        'computed automatically by matplotlib backend.'
    )

    parser.add_argument(
        '--cutout', metavar='SOURCE_IMAGE', type=str, default=None,
        help='path of a FITS image (RGB or grayscale) used to make cutouts '
        'that will be included in the plots. The size of the cutout can be '
        'set using the --cutout-size option.'
    )

    parser.add_argument(
        '--cutout-size', metavar='SIZE', type=str, default='10arcsec',
        help='Set the size of the cutout to %(metavar)s. This option '
        'supports units compatible with astropy (for example "1deg", '
        '"2arcmin", "5arcsec", etc.). If no unit is specified the size is '
        'assumed to be in arcseconds. The default cutout size is %(default)s.'
    )

    parser.add_argument(
        '--redshift', '-z', metavar='Z', type=float, default=None,
        help='Redshift of the objects. If not specified then no lines will be'
        'plotted on the spectra.'
    )

    parser.add_argument(
        '--plot-wrange', metavar='W_RANGE', type=str,
        help='wavelnght range to plot expressed in the form A,B where A and B '
        'are wavelengths in an astropy compatible format (ie. 100angstrom or '
        '1nm or 1). Adimensional values will be considered expressed in '
        'angstroms.'
    )

    parser.add_argument(
        '--cmap', metavar='CMAP', type=str, default='jet',
        help='Set the colormap to use when plotting the spectral datacube '
        'cutout. Can be the name of any valid matplotlib colormap. The '
        'default value is %(metavar)s=%(default)s.'
    )

    parser.add_argument(
        '--outname', '-o', metavar='FILE_NAME', type=str, default='anim.gif',
        help='Set the name of the output file. If not specified, then the '
        'default value %(metavar)s=%(default)s is used.'
    )

    parser.add_argument(
        '--out-dpi', metavar='DPI', type=int, default=75,
        help='Set the DPI of the output file. If not specified, then the '
        'default value %(metavar)s=%(default)s is used.'
    )

    parser.add_argument(
        '--out-fps', metavar='FPS', type=int, default=5,
        help='Set the output framerate to %(metaver)s. The default value is '
        '%(metavar)s=%(default)s'
    )

    parser.add_argument(
        '--out-loop', metavar='LOOP_COUNT', type=int, default=0,
        help='Set the number of times the animation is played. If '
        '%(metavar)s=0 then the animation loops indefinitely. The default '
        'value is %(metavar)s=%(default)s.'
    )

    parser.add_argument(
        '--ffmpeg-exec', metavar='FFMPEG_PATH', type=str, default=FFMPEG_EXC,
        help='Set the path of the FFMPEG executable. If not specified the '
        'ffmepg executable is searched automatically in your system PATH. '
    )

    parser.add_argument(
        '--ffmpeg', action='store_true', default=False,
        help='Force the use the FFMPEG backend instead of imageio, useful if '
        'you want to generate a video instead of a GIF.'
    )

    parser.add_argument(
        'datacube', metavar='DATA_CUBE', type=str, help='The datacube used '
        'to extract the spectra.'
    )

    parser.add_argument(
        'wavelength', metavar='WAVE', type=str, help='centra wavelnght of '
        'the feature to plot expressed in an astropy compatible format (ie. '
        '100angstrom or 1nm or 1). Adimensional values will be considered '
        'expressed in angstroms.'
    )

    parser.add_argument(
        'size', metavar='SIZE', type=str, help='width of the area used to '
        'generate the animation, expressed in an astropy compatible format '
        '(ie. 100angstrom or 1nm or 1). Adimensional values will be considered'
        ' expressed in angstroms.'
    )

    parser.add_argument(
        'spectra', metavar='SPECTRUM', type=str, nargs='+',
        help='Input spectra extracted with specex.'
    )

    args = None
    if options is None:
        args = parser.parse_args()
    else:
        args = parser.parse_args(options)

    return args


def plot_cube_slice_animation(options=None):
    """
    Plot animanted gif of the extracted sources.

    Parameters
    ----------
    options : list of str, optional
        DESCRIPTION. The default is None.

    Returns
    -------
    None.

    """
    global HAS_IMAGEIO
    global FFMPEG_EXC

    if not HAS_IMAGEIO:
        if FFMPEG_EXC is None:
            print(
                "\nERROR: ffmpeg not found and imageio not installed!\n\n"
                "This program requires either the python package imageio (the "
                "preferred option) or a working version of FFMPEG. "
                "If you already have FFMPEG try to manually set the executable"
                " path with the --ffmpeg-exec option\n"
            )
            sys.exit(1)

    quantity_support()
    args = __plot_slice_argshandler(options)

    n_spectra = len(args.spectra)

    cutout_size = apu.Quantity(args.cutout_size)

    out_fps = abs(args.out_fps)
    out_loop = abs(args.out_loop)
    ffmpeg_exec = args.ffmpeg_exec

    if args.ffmpeg:
        HAS_IMAGEIO = False

    if args.cube_vlim is not None:
        cube_vlim = [float(x) for x in args.cube_vlim.split(',')]
        cube_vmin = np.min(cube_vlim)
        cube_vmax = np.max(cube_vlim)
    else:
        cube_vmin = None
        cube_vmax = None

    if args.cutout is not None:
        try:
            big_image = load_rgb_fits(args.cutout)
        except FileNotFoundError:
            print(f"ERROR: file not found '{args.cutout}'")
            sys.exit(1)

        if big_image is None:
            big_image = {
                'data': fits.getdata(args.cutout),
                'wcs': wcs.WCS(fits.getheader(args.cutout)),
                'type': 'gray'
            }
    else:
        big_image = None

    print("Loading spectral data...")
    spectra_list = []
    for j, spectrum_fits_file in enumerate(args.spectra):
        progress = (j + 1) / len(args.spectra)
        sys.stdout.write(f"\r{get_pbar(progress)} {progress:.2%}\r")
        sys.stdout.flush()

        with fits.open(spectrum_fits_file) as hdulist:
            main_header = get_hdu(
                hdulist,
                hdu_index=0,
                valid_names=['PRIMARY', 'primary'],
                msg_index_error="WARNING: No Primary HDU",
                exit_on_errors=False
            ).header
            spec_hdu = get_hdu(
                hdulist,
                valid_names=['SPEC', 'spec', 'SPECTRUM', 'spectrum'],
                msg_err_notfound="WARNING: No spectrum HDU",
                exit_on_errors=False
            )
            var_hdu = get_hdu(
                hdulist,
                valid_names=['VAR', 'var', 'VARIANCE', 'variance'],
                msg_err_notfound="WARNING: No variance HDU",
                exit_on_errors=False
            )

            nan_mask_hdu = get_hdu(
                hdulist,
                valid_names=[
                    'NAN_MASK', 'nan_mask',
                    'NANMASK', 'MASK',
                    'nanmask', 'mask'
                ],
                exit_on_errors=False
            )

            if any(x is None for x in [main_header, spec_hdu, var_hdu]):
                print(f"Skipping file '{spectrum_fits_file}'\n")
                continue

            try:
                object_ra = main_header['RA']
                object_dec = main_header['DEC']
                object_id = main_header['ID']
                extraction_mode = main_header['EXT_MODE']
                specex_apertures = [
                    apu.Quantity(x)
                    for x in json.loads(main_header['EXT_APER'])
                ]
            except KeyError:
                print(
                    f"Skipping file with invalid header: {spectrum_fits_file}"
                )
                continue
            else:

                try:
                    object_coord_frame = main_header['FRAME']
                except KeyError:
                    object_coord_frame = 'fk5'

                obj_center = SkyCoord(
                    object_ra, object_dec,
                    unit='deg',
                    frame=object_coord_frame
                )

            try:
                flux_units = spec_hdu.header['BUNIT']
            except KeyError:
                flux_units = None

            try:
                wavelength_units = spec_hdu.header['CUNIT1']
            except KeyError:
                wavelength_units = None

            flux_data = spec_hdu.data.copy()
            spec_wcs = wcs.WCS(spec_hdu.header, fobj=hdulist).copy()
            var_data = var_hdu.data.copy()

            if nan_mask_hdu is not None:
                nan_mask = nan_mask_hdu.data == 1
            else:
                nan_mask = None

            # NOTE: wavelengths must be in Angstrom units
            pixel = np.arange(len(flux_data))
            wavelengths = spec_wcs.pixel_to_world(pixel).Angstrom

            if big_image is not None:
                if big_image['type'] == 'rgb':
                    cutout_dict = get_rgb_cutout(
                        big_image['data'],
                        center=obj_center,
                        size=cutout_size,
                        data_wcs=big_image['wcs']
                    )
                    cutout = np.asarray(cutout_dict['data']).transpose(1, 2, 0)
                    cutout_wcs = cutout_dict['wcs'][0]
                else:
                    cutout_dict = get_gray_cutout(
                        big_image['data'],
                        center=obj_center,
                        size=cutout_size,
                        data_wcs=big_image['wcs']
                    )
                    cutout = np.array(cutout_dict['data'])
                    cutout_wcs = cutout_dict['wcs']
                cutout_vmin = np.nanmin(big_image['data'])
                cutout_vmax = np.nanmax(big_image['data'])
            else:
                cutout = None
                cutout_wcs = None
                cutout_vmin = None
                cutout_vmax = None

            spectra_list.append(
                (
                    object_id,
                    (wavelengths, wavelength_units),
                    (flux_data, flux_units),
                    var_data,
                    nan_mask,
                    obj_center,
                    cutout,
                    cutout_wcs,
                    (cutout_vmin, cutout_vmax),
                    extraction_mode,
                    specex_apertures
                )
            )

    # We assume all the input spectra come from the same cube and have the
    # same dispersion grid
    wavelengths = spectra_list[0][1][0]

    c_wave = apu.Quantity(args.wavelength)
    size = apu.Quantity(args.size)

    print(
        f"Generating animation centered on {c_wave} with a size of {size}..."
    )

    if str(c_wave.unit).strip() != '':
        c_wave = c_wave.to(apu.angstrom).value
    else:
        c_wave = c_wave.value

    if str(size.unit).strip() != '':
        size = size.to(apu.angstrom).value
    else:
        size = size.value

    w_min = c_wave - size/2
    w_max = c_wave + size/2

    play_wav_mask = (wavelengths >= w_min) & (wavelengths <= w_max)
    play_wav_ang = wavelengths[play_wav_mask]

    with fits.open(args.datacube) as cube_hdul:
        cube_spec_hdu = get_hdu(cube_hdul, ['spec', 'data',])
        cube_wcs = wcs.WCS(cube_spec_hdu.header)
        cube_data = cube_spec_hdu.data
        reduced_cube = cube_data[play_wav_mask]

    if args.plot_wrange is None:
        plot_wmin = w_min - 5*size
        plot_wmax = w_max + 5*size
    else:
        plot_w_list = [
            x.to(apu.angstrom).value if x.unit else x.value
            for x in [apu.Quantity(k) for k in args.plot_wrange.split(',')]
        ]
        plot_wmin = np.min(plot_w_list)
        plot_wmax = np.max(plot_w_list)

    # Create the base canvas
    n_spectra = len(spectra_list)

    cell_c_size = 5
    cell_h_frac = 0.3
    cell_h = 5
    cell_w = 5

    fig = plt.figure(figsize=(cell_w * cell_c_size, cell_c_size * n_spectra))
    gs = GridSpec(cell_h*n_spectra, cell_w, figure=fig, hspace=0.5)

    c_dh = int(cell_h * cell_h_frac)

    axs = []
    if big_image:
        cut_cube_idx = -2
    else:
        cut_cube_idx = -1

    for k, spc in enumerate(spectra_list):
        cube_base_cutout = Cutout2D(
            data=reduced_cube[0],
            position=spc[5],
            size=cutout_size,
            wcs=cube_wcs.celestial
        )

        cell_h_k = cell_h*k
        cell_h_k1 = cell_h*(k+1)
        ax_sp = fig.add_subplot(gs[cell_h_k:cell_h_k1-c_dh, :cut_cube_idx])
        ax_vr = fig.add_subplot(
            gs[cell_h_k1-c_dh:cell_h_k1, :cut_cube_idx],
            sharex=ax_sp
        )
        ax_cut_cube = fig.add_subplot(
            gs[cell_h_k:cell_h_k1, cut_cube_idx],
            projection=cube_base_cutout.wcs
        )

        cube_img = ax_cut_cube.imshow(
            cube_base_cutout.data,
            cmap=args.cmap,
            origin='lower',
            aspect='equal',
            vmin=cube_vmin,
            vmax=cube_vmax
        )

        if big_image:
            ax_cut_img = fig.add_subplot(
                gs[cell_h_k:cell_h_k1, -1],
                projection=spc[7]
            )
        else:
            ax_cut_img = None

        plot_spectrum(
            wavelengths=spc[1][0],
            flux=spc[2][0],
            variance=spc[3],
            nan_mask=spc[4],
            redshift=args.redshift,
            wavelengt_units=spc[1][1],
            flux_units=spc[2][1],
            smoothing=args.spec_smoothing,
            cutout=spc[6],
            cutout_wcs=spc[7],
            cutout_vmin=spc[8][0],
            cutout_vmax=spc[8][1],
            wave_range=[plot_wmin, plot_wmax],
            show_legend=False,
            axs=[ax_sp, ax_vr, ax_cut_img, None],
        )

        spec_cursor = ax_sp.axvline(0, color='red', ls='-', lw=2, alpha=0.7)
        var_cursor = ax_vr.axvline(0, color='red', ls='-', lw=2, alpha=0.7)

        row_axs = {
            'spec': ax_sp,
            'spec_cursor': spec_cursor,
            'var': ax_vr,
            'var_cursor': var_cursor,
            'cube_cutout': ax_cut_cube,
            'img_cutout': ax_cut_img,
            'cube_img': cube_img
        }
        ax_cut_cube.axis('off')

        if ax_cut_img:
            # TODO: check condition to invert axis
            ax_cut_img.invert_xaxis()
            ax_cut_img.invert_yaxis()

        axs.append(row_axs)

    bkg = fig.canvas.copy_from_bbox(fig.bbox)

    if HAS_IMAGEIO:
        writer = imageio.get_writer(
            args.outname,
            mode='I',
            loop=out_loop,
            duration=1 / out_fps
        )

    with tempfile.TemporaryDirectory() as tempdir:
        figname = os.path.join(tempdir, 'frame_%06d.png')
        for j, wav in enumerate(play_wav_ang):
            progress = (j + 1) / len(play_wav_ang)
            sys.stdout.write(f"\r{get_pbar(progress)} {progress:.2%}\r")
            sys.stdout.flush()

            fig.canvas.restore_region(bkg)
            for k, (row_axs, spc) in enumerate(zip(axs, spectra_list)):
                row_axs['spec_cursor'].set_xdata([wav, wav])
                row_axs['var_cursor'].set_xdata([wav, wav])

                current_cutout = Cutout2D(
                    data=reduced_cube[j],
                    position=spc[5],
                    size=cutout_size,
                    wcs=cube_wcs.celestial
                )

                if args.cube_smoothing:
                    current_cutout_data = gaussian_filter(
                        current_cutout.data,
                        args.cube_smoothing,
                        mode='constant'
                    )
                else:
                    current_cutout_data = current_cutout.data

                row_axs['cube_img'].set_data(current_cutout_data)

            fig.canvas.blit(fig.bbox)
            fig.canvas.flush_events()
            fig.savefig(
                figname % j,
                bbox_inches='tight',
                dpi=args.out_dpi
            )
            if HAS_IMAGEIO:
                writer.append_data(imageio.v3.imread(figname % j))

        if HAS_IMAGEIO:
            writer.close()
        else:
            print("Encoding with FFMPEG...")
            ffmpeg_cmd = [
                ffmpeg_exec, '-y', '-i', figname,
                '-framerate', str(out_fps),
                '-loop', str(out_loop),
                args.outname
            ]
            result = subprocess.run(
                ffmpeg_cmd,
                stdout=subprocess.PIPE
            )
            print(result)
    plt.close(fig)
